/*****************************************************************************
 *
 * Copyright (c) 2018-19, Eduard Gerlits, Paula Herber, Josef Treus
 * Westfälische Wilhelms-Universität Münster, AG Embedded Systems 
 * All rights reserved.
 *
 * This file is part of SysCSec
 *
 * SysCSec is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * SysCSec is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SysCSec.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Please report any problems or bugs to: sysc.sec@uni-muenster.de
 *
 ****************************************************************************/
#ifndef _XORENCRYPT_H_
#define _XORENCRYPT_H_

#include <iostream>
#include <systemc.h>
#include "commons.h"

/*
This module implements bitwise XOR encryption, as required by the Mutual Authentication Protocol. The XOR encryption is symmetric, i.e., with a second application of the same key, data is decrypted again.
*/

SC_MODULE(XOREncryptionModule){

	// input key
	sc_fifo_in<sc_biguint<160> > key_in;

	// input data
	sc_fifo_in<sc_biguint<160> > data_in;

	// encrypted data
	sc_fifo_out<sc_biguint<160> > data_out; 

	void exe();
	int mode;
	SC_HAS_PROCESS(XOREncryptionModule);
	XOREncryptionModule(sc_module_name name, int mode): sc_module(name){
		this->mode = mode;
		// SystemC-Process
		SC_THREAD(exe);
	}
};

#endif
